""" This module is used to compute different scores once one step of the
simulation is done.

The city is evaluated wrt the following criteria :
* Pollution, How much the energies have poluted for this tick globally (ie.
  the mean of the pollution produced).
* Spendings, Sum of energy cost.
* Citizens' satisfaction (or happiness), Sum of satisfactions of all citizen
  weighted by the importance each citizen give to each criteria (eather,
  lighting).
"""
from itertools import cycle

from scipy.stats import beta as β_distribution
import numpy as np


def compute_scores(people, energies, lights, heaters):
    spendings_score = spendings(energies)
    pollution_score = pollution(energies)

    nuclear = energies.bought_amounts[1] / energies.bought_amounts.sum()
    satisfaction_score = satisfaction(people, lights, heaters, pollution_score, nuclear)
    return {
        "spendings": spendings_score,
        "pollution": pollution_score,
        "happiness": satisfaction_score
    }


def pollution(energies):
    tot = energies.pollution_factors * energies.bought_amounts
    return (tot / energies.bought_amounts.sum()).sum()


def compute_interraction_score(interraction, item):
    return (interraction @ item.T) / (interraction.sum(1) + 1e-5)


def compute_score(scores, αs, βs, loc, scale):
    pdf_score = [β_distribution.pdf(score, α, β, loc=loc, scale=scale)
                    for score, α, β in zip(scores, αs, βs)]
    #try:
    #    print(params, next(scores), pdf_score)
    #except:
    #    print("\n\nComputeScore\n\n", αs, βs, loc, scale, scores, pdf_score)
    return np.mean(pdf_score)


def satisfaction(people, lights, heaters, pollution_score, nuclear):
    pollution_score = compute_score(cycle([pollution_score]), *people.pollution)

    nuclear_score = compute_score(cycle([nuclear]), *people.nuclear)

    lights_inter = compute_interraction_score(people.lights_interraction, lights)
    light_score = compute_score(lights_inter, *people.lights_pref)

    heaters_inter = compute_interraction_score(people.heaters_interraction, heaters)
    heat_score = compute_score(heaters_inter, *people.heater_pref)

    #print("Scores", pollution_score, nuclear_score, light_score, heat_score)

    return pollution_score + nuclear_score + light_score + heat_score


def spendings(energies):
    return (energies.bought_amounts * energies.costs).sum()



""" Controller for the `Simulator` class.
"""


class Engine:
    def __init__(self, smart_city):
        self.smart_city = smart_city
        self.scores = []

    def __str__(self):
        pass

    def light_up(self, lights):
        self.smart_city.light_up(lights)

    def heat_up(self, heaters):
        self.smart_city.heat_up(heaters)

    def buy_energies(self, amounts):
        self.smart_city.buy_energies(amounts)

    def get_obs(self):
        return {
            "lights": self.smart_city.lights,
            "heaters": self.smart_city.heaters,
            "energies_cost": self.smart_city.energies.costs,
            "energies_amount": self.smart_city.energies.amounts,
            "needed_energy": self.smart_city.needded_energy
        }

    def get_info(self):
        return {
            "light_interraction": self.smart_city.people.lights_interraction,
            "heaters_interraction": self.smart_city.people.heaters_interraction
        }

    def step(self):
        step_scores = self.smart_city.step()
        return step_scores, self.get_obs(), self.get_info()




def estimate_beta_parameters(means, stds, low, high):
    loc = low
    scale = high

    # print("\t\t BEFORE", means, stds)
    means = (means - loc) / scale
    variances = (stds / scale) ** 2
    # print("\t\t AFTER", means, stds / scale)

    temp = means * (1 - means) / variances - 1
    α = means * temp
    β = (1 - means) * temp

    # print(f"""
    #     Estimation of Beta Parameters
    #         α = {α}
    #         β = {β}
    #         loc = {loc}
    #         scale = {scale}

    #     Computed using :
    #         means : {means}
    #         vars : {variances}
    #         loc : {loc}
    #         scale : {scale}
    # """)

    return α, β, loc, scale


class BetaDistributions(object):
    def __init__(self, means, stds, low, high):
        self.alphas, self.betas, self.locs, self.scales = estimate_beta_parameters(means, stds, low, high)
        self.means = means
        self.stds = stds
        assert (self.alphas < 0).sum() == 0
        assert (self.betas < 0).sum() == 0

    def __iter__(self):
        return iter((self.alphas, self.betas, self.means, self.stds))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="smartcity",
    version="1.0.0",
    author="Adrien Pouyet (Ricocotam)",
    author_email="ricocotam@gmail.Com",
    description="A Smart City Env designed for Reinforcement Learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ricocotam/SmartCity-simulator",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    include_package_data=True
)
